#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Mixins are used to ensure capabilities are present and implemented
"""


class Mixin(object):
    pass


class SMUProviderMixin(Mixin):
    def __init__(self, *args, **kwargs):
        super(SMUProviderMixin, self).__init__(*args, **kwargs)
        print("SMU provider started")


class VSenseProviderMixin(Mixin):
    def __init__(self, *args, **kwargs):
        super(VSenseProviderMixin, self).__init__(*args, **kwargs)
        print("VSense provider started")


class X100(SMUProviderMixin, VSenseProviderMixin):
    def __init__(self, *args, **kwargs):
        super(X100, self).__init__(*args, **kwargs)
        print("X100 started")



from xtralien import Device

# Constants
EXT = 1 << 4
FAN = 1 << 5
SHUTTER = 1 << 6
SHORT = 1 << 7


class SmartBoard(object):
    def __init__(self, device):
        """
        The SmartBoard is a wrapper around the Device class that uses
        predefined commands to make using the smartboard easier.
        """
        self.device = device
        self.relays = 1

        self.device.smartio.set.value.port[2](relays, response=0)

    def pixel(pixel, on=True):
        pixel = pixel or 1
        self.device.smartio.set.value.pin[15 + pixel](on, response=0)

    def pixels(*pixels):
        value = 0

        if pixels:
            for pixel in pixels:
                value |= 1 << pixel - 1 if pixel > 0 else 0
        else:
            value = 0b11111111

        self.device.smartio.set.value.port(0, value, response=0)

    def toggle(*toggles):
        self.relays ^= switch
        print(relays)
        self.device.smartio.set.value.port(2, self.relays, response=0)
        sleep(0.1)

    @staticmethod
    def USB(port):
        device = X100.USB(port)
        return SmartBoard(device)

    @staticmethod
    def Network(ip):
        device = Device.Network(ip)
        return SmartBoard(device)

relays = 1



# pylint: disable=W0611,W0614,W0401,W0622,W0404
"""Xtralien Imports

This module is a list of imports used by Xtralien Python Distribution.

Most of these are common functions, however we found that we were using
them a lot so we included them here.
"""

import os  # noqa: F401
import sys  # noqa: F401
from glob import glob  # noqa: F401
import re  # noqa: F401
import csv

import base64  # noqa: F401
import base64 as b64  # noqa: F401
import decimal  # noqa: F401

# Time
from time import *  # noqa: F401

# Warnings
from warnings import *  # noqa: F403

# Xtralien
import xtralien  # noqa: F401
from xtralien import X100, Device  # noqa: F401
from xtralien.serial_utils import serial_ports  # noqa: F401

# Scipy
import scipy  # noqa: F401
from scipy import *  # noqa: F401

# Numpy
import numpy as np  # noqa: F401
from numpy import *  # noqa: F401
from numpy.random import *  # noqa: F401

# Matplotlib
import matplotlib.pyplot as plt  # noqa: F401
from matplotlib.pyplot import *  # noqa: F401

# 3D plotting
from mpl_toolkits.mplot3d import axes3d, Axes3D  # noqa: F401


def print_header(*args):
    """
    Print a header containing the given elements.

    This is just a specific format for printing.
    """
    for arg in args:
        if (not arg) or arg == '':
            buffer_ = ''
        else:
            buffer_ = ' '
        print("{:#^80}".format(arg.join([buffer_] * 2)))


def array_to_csv(arr, fname, *args, **kwargs):
    """
    Save an array to a csv, with a given filename.

    The array should be a numpy array.
    """
    try:
        savetxt(fname, arr, *args, delimiter=",", **kwargs)
    except (IOError, ValueError):
        savetxt(fname, arr, *args, delimiter=",", fmt="%s", **kwargs)


def load_csv(fname, skip_headers=False, *args, **kwargs):
    """
    Load a csv file from a given filename.

    @return Numpy array
    """
    arr = None

    with open(fname, 'r+') as fileh:
        if skip_headers:
            fileh.readline()

        try:
            arr = loadtxt(fileh, *args, delimiter=',', **kwargs)  # noqa: F405
        except (IOError, ValueError):
            import csv
            reader = csv.reader(fileh, delimiter=',')
            arr = []
            for row in reader:
                rowdata = []
                for data in row:  # Datapoint in row
                    try:
                        rowdata.append(float(data))
                    except ValueError:
                        rowdata.append(data)
                    arr.append(rowdata)
    return arr

# Delete some constants
del altzone  # noqa: F821
del daylight  # noqa: F821
del timezone  # noqa: F821
del tzname  # noqa: F821
euler_gamma = None  # noqa: F821 # pylint: disable=locally-disabled,invalid-name
del euler_gamma  # noqa: F821



MONITOR = 0b001
TRIGGER = 0b010


class Experiment(object):
    def __init__(self, name=None, description=None, mode=MONITOR, frequency=1, devices=None):
        self.devices = devices
        self.frequency = frequency
        self.mode = mode

    def __enter__(self):
        pass

    def __exit__(self, *args):
        self.save()



import sys
import glob
import serial


def serial_ports():
    """Lists serial ports

    :raises EnvironmentError:
        On unsupported or unknown platforms
    :returns:
        A list of available serial ports
    """
    if sys.platform.startswith('win'):
        ports = ['COM' + str(i + 1) for i in range(256)]

    elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
        # this is to exclude your current terminal "/dev/tty"
        ports = glob.glob('/dev/tty[A-Za-z0-9]*')

    elif sys.platform.startswith('darwin'):
        ports = glob.glob('/dev/tty.usb*')

    else:
        raise EnvironmentError('Unsupported platform')

    result = []
    for port in ports:
        try:
            s = serial.Serial(port, timeout=0.01)
            s.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass
    return result


if __name__ == '__main__':
    print(serial_ports())



from xtralien.imports import *  # noqa: F403


def prompt(vars=None, message=""):
    prompt_message = message
    try:
        from IPython.Shell import IPShellEmbed
        ipshell = IPShellEmbed(
            argv=[''],
            banner=prompt_message,
            exit_msg="Exiting..."
        )
        return ipshell()
    except ImportError:
        if vars is None:
            vars = globals()
        import code
        import rlcompleter  # noqa: F401
        import readline
        readline.parse_and_bind("tab: complete")
        print(prompt_message)
        shell = code.InteractiveConsole(vars)
        return shell.interact()

prompt_ = prompt()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Xtralien
"""
# Create a basic logger to make logging easier
import logging
import os
import time
import re
import threading
import random
import datetime
import sys
import socket

from xtralien.serial_utils import serial_ports

log_levels = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warn': logging.WARNING,
    'error': logging.ERROR
}

logging.basicConfig()
logger = logging.getLogger('Xtralien')
logger.setLevel(
    log_levels.get(os.getenv('LOG', 'warn').lower(), logging.WARNING)
)

try:
    import numpy
except ImportError:
    logger.warn("Numpy not found, narray and nmatrix will fail")


if sys.version_info.major < 3:
    logger.warn("Module not supported on Python 2.x")

# Try and import the serial module (supports USB-serial communication)
try:
    import serial
except ImportError:
    serial = None
    logger.warn("The serial module was not found, USB not supported")


def process_strip(x):
    return x.strip('\n[];')


def process_array(x):
    data = [float(y) for y in x.strip('\n[];').split(';')]
    try:
        return numpy.array(data)
    except NameError:
        return data


def process_matrix(x):
    data = [
        [float(z) for z in y.split(',')]
        for y in x.strip('\n[];').split(';')
    ]
    try:
        return numpy.array(data)
    except NameError:
        return data


number_regex = r"(\-|\+)?[0-9]+(\.[0-9]+)?(e-?[0-9]+(\.[0-9]+)?)?"
re_matrix = re.compile(
    r'(\[({number},{number}(;?))+\])\n?'.format(number=number_regex)
)
re_array = re.compile(r'(\[({number}(;?))+\])\n?'.format(number=number_regex))
re_number = re.compile(r'{number}\n?'.format(number=number_regex))


def process_auto(x=None):
    if x is None:
        return x
    if re_matrix.fullmatch(x):
        return process_matrix(x)
    elif re_array.fullmatch(x):
        return process_array(x)
    elif re_number.fullmatch(x):
        return float(x)
    elif '\n' in x:
        split_string = x.strip('\n;[]').split('\n')
        if len(split_string) < 2:
            return split_string[0]
        return split_string
    else:
        return x


class Device(object):
    connections = []
    current_selection = []
    formatters = {
        'strip': process_strip,
        'array': process_array,
        'matrix': process_matrix,
        'number': lambda x: float(x),
        'none': lambda x: x,
        'auto': process_auto
    }

    def __init__(self, addr=None, port=None, serial_timeout=0.1):
        self.connections = []
        self.current_selection = []
        self.in_progress = False

        if port:
            self.add_connection(SocketConnection(addr, port))
        elif addr:
            self.add_connection(SerialConnection(addr, timeout=serial_timeout))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if traceback is not None:
            print(traceback)
        self.close()

    @property
    def connection(self):
        return self.connections[0]

    def add_connection(self, connection):
        self.connections.append(connection)

    def command(self, command, returns=False, sleep_time=0.001):
        if sleep_time is not None:
            time.sleep(sleep_time)
        if self.connections == []:
            logger.error(
                "Can't send command '{cmd}'\
                because there are no open connections'".format(
                    cmd=command
                )
            )
        for conn in self.connections:
            if True:  # try:
                conn.write(command)
                return conn.read(returns)
            else:  # except (Exception, e):
                # print(e)
                conn.close()
                continue
        logger.error(
            "Can't send command '{cmd}'\
            because there are no open connections".format(
                cmd=command
            )
        )

    def close(self):
        for conn in self.connections:
            conn.close()

    @property
    def serial(self):
        _serial = int('0x' + self("serial", format=None), 16)
        return {
            # 16 bits
            "board_number": (_serial & 0x00000000FFFF),
            # 6 bits
            "week": (_serial & 0x0000003F0000) >> 16,
            # 8 bits
            "year": (_serial & 0x00003FC00000) >> 22,
            # 8 bits
            "model": (_serial & 0x003FC0000000) >> 30,
            # 10 bits
            "product": (_serial & 0xFFC000000000) >> 38
        }

    @serial.setter
    def serial(self, serial_dict):
        # Set defaults
        dt = datetime.datetime.now()
        week = serial_dict.get("week", int(dt.strftime("%W")))
        year = serial_dict.get("year", dt.year - 2000)
        model = serial_dict.get('model', 0)
        product = serial_dict.get('product', 0)
        board = serial_dict.get('board_number', 0)
        # Create Serial
        _serial = 0x000000000000
        _serial |= board & 0xffff
        _serial |= (week & 0x3f) << 16
        _serial |= (year & 0xff) << 22
        _serial |= (model & 0xff) << 30
        _serial |= (product & 0x3ff) << 38

        for i in range(6):
            self.eeprom.set(16383-i, (_serial & 0xff), response=0)
            _serial >>= 8
            time.sleep(0.1)

        return self.serial

    def __getattribute__(self, x):
        if '__' in x or x in object.__dir__(self):
            return object.__getattribute__(self, x)
        else:
            self.current_selection.append(x)
            return self

    def __getitem__(self, x):
        self.current_selection.append(x)
        return self

    def __call__(self, *args, **kwargs):
        sleep_time = kwargs.get("sleep_time", 0.001)
        self.current_selection += args
        returns = kwargs.get('response', True) or kwargs.get('callback', False)
        command = ' '.join([str(x) for x in self.current_selection])
        self.current_selection = []

        def formatter(x):
            return x

        if returns:
            try:
                formatter = self.formatters.get(
                    kwargs.get('format', 'auto'),
                    formatter
                )
            except KeyError:
                if kwargs.get('format', None):
                    logger.warning('Formatter not found')

        callback = kwargs.get('callback', None)

        if callback:
            def async_function():
                while self.in_progress:
                    continue
                self.in_progress = True
                data = formatter(self.command(command, returns=True))
                self.in_progress = False
                callback(data)
            return threading.Thread(target=async_function).start()

        self.in_progress = True
        data = formatter(
            self.command(command, returns=returns, sleep_time=sleep_time)
        )
        self.in_progress = False
        return data

    def __repr__(self):
        if len(self.connections):
            return "<Device connection={connection}/>".format(
                connection=self.connections[0]
            )
        else:
            return "<Device connection=None/>"

    @staticmethod
    def discover(broadcast_address=None, timeout=0.1, *args, **kwargs):
        broadcast_address = broadcast_address or '<broadcast>'
        udp_socket = socket.socket(
            family=socket.AF_INET,
            type=socket.SOCK_DGRAM
        )
        udp_socket.bind(('0.0.0.0', random.randrange(6000, 50000)))
        udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        udp_socket.settimeout(timeout)
        udp_socket.sendto(b"xtra", (broadcast_address, 8889))
        devices = []
        try:
            start_time = time.time()
            while True:
                if time.time() > (start_time + timeout):
                    break
                (_, ipaddr) = udp_socket.recvfrom(4)
                if ipaddr:
                    start_time = time.time()
                devices.append(Device(addr=ipaddr[0], port=8888))
        except socket.timeout:
            pass
        finally:
            udp_socket.close()
        return devices

    def dup(self):
        return DeviceDuplicate(self)

    @staticmethod
    def USB(com=None, *args, **kwargs):
        if com is None:
            while True:
                try:
                    com = serial_ports()[0]
                except IndexError:
                    continue
                else:
                    break

        return Device(com, *args, **kwargs)

    fromUSB = USB
    openUSB = USB

    @staticmethod
    def COM(com, *args, **kwargs):
        return Device.USB("COM{}".format(com), *args, **kwargs)

    fromCOM = COM
    openCOM = COM

    @staticmethod
    def Network(ip, *args, **kwargs):
        return Device(ip, 8888, *args, **kwargs)

    fromNetwork = Network
    openNetwork = Network

    @staticmethod
    def first(*args, **kwargs):
        while True:
            try:
                try:
                    com = serial_ports()[0]
                    return Device.USB(com, *args, **kwargs)
                except IndexError:
                    return Device.discover(*args, **kwargs)[0]
            except IndexError:
                continue


class DeviceDuplicate(object):
    def __init__(self, device):
        self.device = device
        self.command_base = self.device.current_selection
        self.device.current_selection = []
        self.current_selection = []

    def __getattribute__(self, x):
        if '__' in x or x in object.__dir__(self):
            return object.__getattribute__(self, x)
        else:
            self.current_selection.append(x)
            return self

    def __getitem__(self, x):
        self.current_selection.append(x)
        return self

    def __call__(self, *args, **kwargs):
        self.device.current_selection = []
        response = self.device(
            ' '.join(self.command_base + self.current_selection),
            *args,
            **kwargs
        )
        self.current_selection = []
        return response


class Connection(object):
    def __init__(self):
        pass

    def read(self, *args, **kwargs):
        logging.error("Method not implemented (%s)" % self.read)

    def write(self, *args, **kwargs):
        logging.error("Method not implemented (%s)" % self.write)

    def close(self, *args, **kwargs):
        logging.error("Method not implemented (%s)" % self.close)


class SocketConnection(Connection):
    def __init__(self, host, port, timeout=0.07):
        super(SocketConnection, self).__init__()
        self.host = host
        self.port = port
        self.socket = socket.socket()
        self.socket.connect((host, port))
        self.socket.settimeout(timeout)

    def read(self, wait=True):
        retval = ""
        if wait:
            while retval == "":
                try:
                    retval = retval + str(self.socket.recv(576), 'utf-8')
                except socket.timeout:
                    continue

        while True:
            try:
                retval = retval + str(self.socket.recv(576), 'utf-8')
            except socket.timeout:
                break

        return retval

    def write(self, cmd):
        if type(cmd) == str:
            cmd = bytes(cmd, 'utf-8')
        self.socket.send(cmd)

    def close(self):
        self.socket.close()

    def __repr__(self):
        return "<Socket {host}:{port} />".format(
            host=self.host,
            port=self.port
        )


class SerialConnection(Connection):
    def __init__(self, port, timeout=0.1):
        super(SerialConnection, self).__init__()
        self.port = port
        self.connection = serial.Serial(port, timeout=timeout)

    def read(self, wait=True):
        retval = ""

        while wait and (self.connection.inWaiting() == 0):
            continue

        while self.connection.inWaiting() > 0:
            try:
                retval = retval + str(
                    self.connection.read(
                        self.connection.inWaiting()
                    ),
                    'utf-8'
                )
            except Exception:
                break

        return retval

    def write(self, cmd):
        if type(cmd) == str:
            cmd = bytes(cmd, 'utf-8')
        self.connection.write(cmd)
        while self.connection.out_waiting > 0:
            continue

    def close(self):
        self.connection.close()

    def __repr__(self):
        return "<Serial/USB {connection} />".format(connection=self.port)


X100 = Device



from setuptools import setup

try:
    with open('README.md') as f:
        long_description = f.read()
except:
    long_description = ''

setup(
    name='xtralien',
    version='2.9.1',
    description='A connector to implement connecting to CLOI-based instruments',
    long_description=long_description,
    author='Xtralien',
    author_email='jack@xtralien.com',
    url='https://github.com/xtralien/pyxtralien.git',
    packages=['xtralien'],
    extras_require={
        'Serial': ['pyserial'],
        'keithley': ['pyVISA'],
        'prompt': ['pyreadline']
    },
    license='GPLv3'
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Print a list of all devices discovered on the network
"""

import xtralien


def main():
    """
    Attempt to discover all devices on the network and print any
    found.
    """
    for dev in xtralien.Device.discover(timeout=1.0):
        print(dev.connections[0].host)


if __name__ == "__main__":
    main()
